/* ====================================================================
   file    : wingetprinters.p
   purpose : function WINGetPrinters returns a list of printers
   by      : Brad Long <bjl@uq.net.au>
   dd      : November 5, 1999
   based on the EnumPrinters example on web-page
       http://www.global-shared.com/api/prienum.html
   ==================================================================== */

/********************** GET VERSION ***********************************/

procedure GetVersionExA external "kernel32":
    define input parameter v_version-info as long.
    define return parameter v_return-value as long.
end procedure.

FUNCTION WINGetVersion
RETURNS CHARACTER
    (
    ):
/*-----------------------------------------------------------------------------
  Purpose: Calls the WINAPI function GetVersionExA to determine the version
           of the Windows operating system that is running on the machine.
    Notes: Returns "95" for Windows 95, "98" for Windows 98, "NT" for Windows NT
           Returns "undef" if unable to determine platform.
------------------------------------------------------------------------------*/

    def var v_version-buf as memptr.
    def var v_platform-id as integer no-undo.
    def var v_platform-desc as char no-undo.
    def var v_major-version as integer no-undo.
    def var v_minor-version as integer no-undo.
    def var v_return-value  as integer no-undo.

    set-size(v_version-buf)   = 148.
    put-long(v_version-buf,1) = 148.

    run GetVersionExA (input get-pointer-value(v_version-buf),
                       output v_return-value).

    v_platform-id = get-long(v_version-buf,17).

    case v_platform-id:
        when 1 then do:
            v_minor-version = get-byte(v_version-buf,15).
            v_major-version = get-byte(v_version-buf,16).
        end.
        otherwise do:
            v_major-version = get-long(v_version-buf,5).
            v_minor-version = get-long(v_version-buf,9).
        end.
    end.

    case v_platform-id:
        when 0 then v_platform-desc = "3.1".
        when 1 then
        do:
            if v_minor-version eq 0 then v_platform-desc = "95".
            else if v_minor-version gt 0 then v_platform-desc = "98".
            else v_platform-desc = "undef".
        end.
        when 2 then
            v_platform-desc = "NT".
        otherwise
            v_platform-desc = "undef".
    end.

    set-size(v_version-buf) = 0.

    return v_platform-desc.

END FUNCTION.


/*********************** GET PRINTERS *********************************/

procedure EnumPrintersA external "winspool.drv":
    define input parameter v_flags as long.
    define input parameter v_name as char.
    define input parameter v_level as long.
    define input parameter v_printer-info as long.
    define input parameter v_num-bytes as long.
    define output parameter v_needed as long.
    define output parameter v_recs_returned as long.
    define return parameter v_return-value as long.
end procedure.

/* WINAPI printer constants */
&global-define PRINTER_ENUM_DEFAULT      1
&global-define PRINTER_ENUM_LOCAL        2
&global-define PRINTER_ENUM_CONNECTIONS  4
&global-define PRINTER_ENUM_FAVORITE     4
&global-define PRINTER_ENUM_NAME         8
&global-define PRINTER_ENUM_REMOTE       16
&global-define PRINTER_ENUM_SHARED       32
&global-define PRINTER_ENUM_NETWORK      64
&global-define PRINTER_ENUM_EXPAND       16384
&global-define PRINTER_ENUM_CONTAINER    32768
&global-define PRINTER_ENUM_ICONMASK     16711680
&global-define PRINTER_ENUM_ICON1        65536
&global-define PRINTER_ENUM_ICON2        131070
&global-define PRINTER_ENUM_ICON3        262144
&global-define PRINTER_ENUM_ICON4        524288
&global-define PRINTER_ENUM_ICON5        1048576
&global-define PRINTER_ENUM_ICON6        2097152
&global-define PRINTER_ENUM_ICON7        4194304
&global-define PRINTER_ENUM_ICON8        8388608

&global-define PRINTER_INFO_1            1
&global-define PRINTER_INFO_2            2
&global-define PRINTER_INFO_3            3
&global-define PRINTER_INFO_4            4
&global-define PRINTER_INFO_5            5
&global-define PRINTER_INFO_6            6
&global-define PRINTER_INFO_7            7

FUNCTION WINGetPrinters
RETURNS CHARACTER
    (
        input v_search as logical
    ):
/*------------------------------------------------------------------------------
  Purpose: makes a WIN API call to EnumPrintersA to determine available printers
    Notes: if v_search is true, the network is searched for all available
           printers. Otherwise, the printers registered with the PC are returned.
           In both cases, local and network printers are returned. The difference is
           that when v_search is true, printers are opened for availability. So it
           checks all printers local or network that are currently available. If your
           environment is quite stable or you only care about the printers that are
           setup in the "Printers" group on the PC, use v_search = no. Searching the
           network with v_search = yes apparently takes longer.
           I didn't notice any performance degradation, but that may be because all
           printers were up and running. Maybe a slight stall when the printer
           is down, but haven't checked this.
           The return value is a comma separated list of printer names.
------------------------------------------------------------------------------*/

    define variable v_printer-enum as memptr no-undo.
    define variable v_num-bytes as integer no-undo.
    define variable v_recs-returned as integer no-undo.
    define variable v_return-value as integer no-undo.

    define variable v_printer-info as memptr no-undo.
    define variable v_struct-size as integer no-undo.

    define variable v_i as integer no-undo.
    define variable v_printer-name as memptr no-undo.
    define variable v_server-name as memptr no-undo.

    define variable v_win-version as char no-undo.
    define variable v_domain as char no-undo.
    define variable v_flag as integer no-undo.

    define variable v_printer-list as char no-undo.


    v_win-version = WINGetVersion().

    if v_win-version = "undef" then return "".

    set-size(v_printer-enum) = 30.  /* set a default value */

    /* Windows 3.1, 95 or 98 return all local & network printers anyway */
    if lookup(v_win-version, "3.1,95,98") > 0 then
    do:
        if v_search then
        do:
            v_struct-size = 84.

            /* The first call to EnumPrinters is only to get the required memory size */
            run EnumPrintersA({&PRINTER_ENUM_LOCAL},
                              "",
                              {&PRINTER_INFO_2},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

            /* v_return-value will now be FALSE (=error) because we did not
               supply enough memory. But at least we know now how much
               memory was required (v_num-bytes) and also how many printers
               were found (v_recs-returned) */

            /* no printers installed, then return (rare) */
            if v_num-bytes = 0 then
            do:
                set-size(v_printer-enum) = 0.
                return "".
            end.

            /* Reset the size of v_printer-enum to the correct size */
            set-size(v_printer-enum) = 0.
            set-size(v_printer-enum) = v_num-bytes.

            /* The second call actually fills the v_printer-enum structure */
            run EnumPrintersA({&PRINTER_ENUM_LOCAL},
                              "",
                              {&PRINTER_INFO_2},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

            /* v_printer-enum holds v_recs-returned PRINTER_INFO_2 records.
               The number of bytes copied to v_printer-enum is v_num-bytes.  */
            do v_i = 0 to v_recs-returned - 1:

                set-pointer-value(v_printer-info) = get-pointer-value(v_printer-enum) + (v_i * v_struct-size).
                set-pointer-value(v_printer-name) = get-long(v_printer-info,5).
                v_printer-list = v_printer-list + "," + get-string(v_printer-name,1).
            end.

        end. /* if v_search */

        else do:  /* not v_search */

             v_struct-size = 20.    /* the size of a PRINTER_INFO_5 record */

            /* The first call to EnumPrinters is only to get the required memory size */
            run EnumPrintersA({&PRINTER_ENUM_LOCAL},
                              "",
                              {&PRINTER_INFO_5},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

            /* v_return-value will now be FALSE (=error) because we did not
               supply enough memory. But at least we know now how much
               memory was required (v_num-bytes) and also how many printers
               were found (v_recs-returned) */

            /* no printers installed, then return (rare) */
            if v_num-bytes = 0 then
            do:
                set-size(v_printer-enum) = 0.
                return "".
            end.

            /* Reset the size of v_printer-enum to the correct size */
            set-size(v_printer-enum) = 0.
            set-size(v_printer-enum) = v_num-bytes.

            /* The second call actually fills the v_printer-enum structure */
            run EnumPrintersA({&PRINTER_ENUM_LOCAL},
                              "",
                              {&PRINTER_INFO_5},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

            /* v_printer-enum holds v_recs-returned PRINTER_INFO_5 records.
               The number of bytes copied to v_printer-enum is v_num-bytes. */
            do v_i = 0 to v_recs-returned - 1:

                set-pointer-value(v_printer-info) = get-pointer-value(v_printer-enum) + (v_i * v_struct-size).
                set-pointer-value(v_printer-name) = get-long(v_printer-info,1).
                v_printer-list = v_printer-list + "," + get-string(v_printer-name,1).
            end.

        end.  /* not v_search */

    end.  /* win 3.1, 95, 98 */

    else if v_win-version = "NT" then
    do:

        if not v_search then  /* get printers that have been set up for this PC - NT only */
        do:

            v_struct-size = 12.      /* The size of one PRINTER_INFO_4 record */

            /* The first call merely gets the required size for v_printer-enum */
            run EnumPrintersA({&PRINTER_ENUM_CONNECTIONS},
                          "",
                          {&PRINTER_INFO_4},
                          get-pointer-value(v_printer-enum),
                          get-size(v_printer-enum),
                          output v_num-bytes,
                          output v_recs-returned,
                          output v_return-value).

            /* v_return-value will now be FALSE (=error) because we did not
               supply enough memory. But at least we know now how much
               memory was required (v_num-bytes) and also how many printers
               were found (v_recs-returned) */

            /* no printers installed, then return (rare) */
            if v_num-bytes = 0 then
            do:
                set-size(v_printer-enum) = 0.
                return "".
            end.

            /* Reset the size of v_printer-enum to the correct size */
            assign
            set-size(v_printer-enum) = 0
            set-size(v_printer-enum) = v_num-bytes.

            /* The second call actually fills the v_printer-enum structure */

            run EnumPrintersA({&PRINTER_ENUM_CONNECTIONS},
                          "",
                          {&PRINTER_INFO_4},
                          get-pointer-value(v_printer-enum),
                          get-size(v_printer-enum),
                          output v_num-bytes,
                          output v_recs-returned,
                          output v_return-value).

            /* v_printer-enum holds v-recs-returned number of PRINTER_INFO_4 structures.
               The number of bytes copied to v_printer-enum is v_num-bytes. */

            do v_i = 0 to v_recs-returned - 1:

                set-pointer-value(v_printer-info) = get-pointer-value(v_printer-enum) + (v_i * v_struct-size).
                set-pointer-value(v_printer-name) = get-long(v_printer-info,1).
                v_printer-list = v_printer-list + "," + get-string(v_printer-name,1).

            end.

        end. /* if not v_search */

        else    /* get all printers available to this PCs domain - NT only */
        do:
            /* The first call to EnumPrinters is only to get the required memory size */

            v_struct-size = 16. /* the size of a PRINTER_INFO_1 record */

            run EnumPrintersA({&PRINTER_ENUM_REMOTE},
                          "",
                          {&PRINTER_INFO_1},
                          get-pointer-value(v_printer-enum),
                          get-size(v_printer-enum),
                          output v_num-bytes,
                          output v_recs-returned,
                          output v_return-value).

            /* v_return-value will now be FALSE (=error) because we did not
               supply enough memory. But at least we know now how much
               memory was required (v_num-bytes) and also how many domains
               were found (v_recs-returned) */

            /* no domains available, then return (rare) */
            if v_num-bytes = 0 then
            do:
                set-size(v_printer-enum) = 0.
                return "".
            end.

            /* Reset the size of v_printer-enum to the correct size */
            assign
            set-size(v_printer-enum) = 0
            set-size(v_printer-enum) = v_num-bytes.

            /* The second call actually fills the v_printer-enum structure */
            /* This gets the PC's domain / print servers */
            run EnumPrintersA({&PRINTER_ENUM_REMOTE},
                          "",
                          {&PRINTER_INFO_1},
                          get-pointer-value(v_printer-enum),
                          get-size(v_printer-enum),
                          output v_num-bytes,
                          output v_recs-returned,
                          output v_return-value).

            /* v_printer-enum holds a PRINTER_INFO_1 record.
               the number of records is v_recs-returned.
               the number of bytes copied to v_printer-enum is v_num-bytes. */

            /* gets DOMAIN info (not printer info) */
            do v_i = 0 to v_recs-returned - 1:

                set-pointer-value(v_printer-info) = get-pointer-value(v_printer-enum) + (v_i * v_struct-size).
                v_flag = get-long(v_printer-info, 1).
                set-pointer-value(v_printer-name) = get-long(v_printer-info, 9).
                v_domain = get-string(v_printer-name, 1).
                leave. /* should only return 1 record for this PC's domain */
            end.

            if true then
            do:
                set-size(v_printer-enum) = 30.  /* set a default value */

                run EnumPrintersA({&PRINTER_ENUM_NAME},
                              v_domain,
                              {&PRINTER_INFO_1},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

                /* v_return-value will now be FALSE (=error) because we did not
                   supply enough memory. But at least we know now how much 
                   memory was required (v_num-bytes) and also how many printers
                   were found (v_recs-returned) */

                /* no printers installed, then return (rare) */
                if v_num-bytes = 0 then
                do:
                    set-size(v_printer-enum) = 0.
                    return "".
                end.

                /* Reset the size of v_printer-enum to the correct size */
                assign
                set-size(v_printer-enum) = 0
                set-size(v_printer-enum) = v_num-bytes.

                /* The second call actually fills the v_printer-enum structure */

                run EnumPrintersA({&PRINTER_ENUM_NAME},
                              v_domain,
                              {&PRINTER_INFO_1},
                              get-pointer-value(v_printer-enum),
                              get-size(v_printer-enum),
                              output v_num-bytes,
                              output v_recs-returned,
                              output v_return-value).

                /* v_printer-enum holds v_recs-returned PRINTER_INFO_1 records.
                   the number of bytes copied to v_printer-enum is v_num-bytes. */

                do v_i = 0 to v_recs-returned - 1:

                    set-pointer-value(v_printer-info) = get-pointer-value(v_printer-enum) + (v_i * v_struct-size).
                    set-pointer-value(v_printer-name) = get-long(v_printer-info,9).
                    v_printer-list = v_printer-list + "," + get-string(v_printer-name,1).
                end.
            end.   /* if true */
        end.  /* get all available/remote printers for NT */
    end.  /* NT */

    if length(v_printer-list) > 0 then 
       v_printer-list = substring(v_printer-list, 2).

    set-size(v_printer-enum) = 0.
    return v_printer-list.

END FUNCTION.


